unit gNutella;

//basic gnutella transport encapsulation

(******************************************************************
Copyright (C)2000, Michael Ax. Created 2000. All rights reserved.
this code is based on the initial (pre-colored) post of this unit
published at http://gnutella.wego.com ~4/10/2000. I've merged
the copyright notice used on the web on 4/21/2000 even though none
was present in the original work. changes below:

max v.1 Includes a few changes to the interface of the unit which
make it more suitable for inheritance and outside use. Altered the
way .log() is used internally to make the code more useful in
conjunction with the itty bitty 'view' client is now includes.
Changed an event type to use Speed:Integer (no longer WORD).. and
defined and event type for the post-notification.

this work's a 'somewhat applied' version of the TGnuTellaTrans
component. call it what you will and enjoy the flexibility of
this protocol and the potential that it offers for cross-platform
application design and development.  -- enjoy!  ax@sonic.net

i would like to see the protocol have more intelligence in each
connection so that connections dropping and conectee ttl spoofing
are identifyable.

(******************************************************************
Copyright (C) 2000, Bryan Mayland Created 2000. All rights reserved.
License to copy and use this software is granted provided that it
is identified as the "CapnBry gnutella transport algorithm" in all
material mentioning or referencing this software or this function.
License is also granted to make and use derivative works provided
that such works are identified as "derived from the CapnBry gnutella
transport algorithm" in all material mentioning or referencing
the derived work. Bryan Mayland makes no representations concerning
either the merchantability of this software or the suitability of
this software for any particular purpose. It is provided "as is"
without express or implied warranty of any kind. These notices must
be retained in any copies of any part of this documentation and/or
software.
*******************************************************************)

interface

uses
  Windows, WinSock, SysUtils, Classes, ScktComp;

type
  Pgnutella_header = ^gnutella_header;
  gnutella_header = packed record
    MsgID:    TGUID;
    Func:     byte;
    TTLRemaining:   byte;
    HopsTaken:      byte;
    DataLen:        integer;
  end;

  Pgnutella_init_response = ^gnutella_init_response;
  gnutella_init_response = packed record
    port:     WORD;
    ip:       Integer;
    filecnt:  DWORD;
    totsize:  DWORD;
  end;

  Pgnutella_query_response_hdr = ^gnutella_query_response_hdr;
  gnutella_query_response_hdr = packed record
    nrecs:    BYTE;
    port:     WORD;
    ip:       Integer;
    speed:    Integer;
  end;

  Pgnutella_query_response_rec_hdr = ^gnutella_query_response_rec_hdr;
  gnutella_query_response_rec_hdr = packed record
    FileIdx:  Integer;
    FileSize: Integer;
  end;

  Pgnutella_query_response_ftr = ^gnutella_query_response_ftr;
  gnutella_query_response_ftr = packed record
    clientid128:   TGUID;
  end;

  Pgnutella_push_req = ^gnutella_push_req;
  gnutella_push_req = packed record
    ClientID128:  TGUID;
    FileIndex:    integer;
    RequestIP:    integer;
    RequestPort:  WORD;
  end;

  Pclient_info_rec = ^client_info_rec;
  client_info_rec = record
    Socket:     TCustomWinSocket;
    Incoming:   boolean;
    ProtocolConnected:  boolean;
    Host:       string;
    RecvBuf:    TMemoryStream;
  end;

  TStringEvent = procedure (Sender : TObject; const s : string) of Object;
  TClientSockEvent = procedure (const CInfo : client_info_rec) of Object;
  TInitRespEvent = procedure (const HostIP : string; HostPort : WORD; FileCnt: integer; FileSize: integer) of Object;
  TSearchReqEvent = procedure (MinSpeed : WORD; const Search : string) of Object;
  TSrchRespBeginEvent = procedure (const HostIP : string; HostPort : WORD; Speed : integer) of Object;
  TSrchRespItemEvent = procedure (FileIdx, FileSize, HostSpeed : integer; const FileName : string) of Object;
  TSrchRespEndEvent = procedure (ClientID128 : TGUID) of Object;
  TPushRequest = procedure(Sender:TObject; ppr:Pgnutella_push_req) of Object;

  TGnuTelMessageList = class(TStringList)
  private
  protected
  public
    function Find(const S: string; var Index: Integer): Boolean; override;
    end;

  TGnuTellaTrans = class(TComponent)
  private
    SockList:   TList;
    ServerSock: TServerSocket;
    FGTNAvgPing: integer;
    FSearchesIn: integer;
    FGTNHosts: integer;
    FGTNSize: integer;
    FGTNFiles: integer;
    FMessagesIn: integer;
    FOnLogStatus: TStringEvent;
    FOnClientConnected: TClientSockEvent;
    FOnClientCreated: TClientSockEvent;
    FOnClientProtocolConnect: TClientSockEvent;
    FMyTTL: BYTE;
    FMaxTTL: BYTE;
    FOnInitResponse: TInitRespEvent;
    FOnSearchReq: TSearchReqEvent;
    FOnPushRequest: TPushRequest;
    FOnSrchRespBegin: TSrchRespBeginEvent;
    FOnSrchRespEnd: TSrchRespEndEvent;
    FOnSrchRespItem: TSrchRespItemEvent;
    FOnClientDisconnect: TClientSockEvent;
    FListenPort: integer;
    FLocalIP: string;
    FLocalFileSize: integer;
    FLocalFileCount: integer;
    FSendInitOnConnect: boolean;
    FOnCatchHost: TStringEvent;
    FConnectionSpeed: integer;
    FRouteMsgList: TGnuTelMessageList;

    procedure GetGNUTELLAOK(Socket: TCustomWinSocket);
    procedure GetGNUTELLA_CONNECT(Socket: TCustomWinSocket);
    procedure ReadSocketData(Socket: TCustomWinSocket);
    procedure DealWithData(CInfo : client_info_rec; pData : PChar; DataSize : integer);
    procedure GotNewMessage(const CInfo : client_info_rec);
    procedure hdr_to_text(const hdr: gnutella_header);
    function FuncToStr(Func: BYTE): string;
    procedure getInit(const CInfo : client_info_rec);
    procedure getInitResponse(const CInfo : client_info_rec);
    procedure getQuery(const CInfo : client_info_rec);
    procedure getQueryResponse(const CInfo : client_info_rec);
    procedure getUnknown(const CInfo : client_info_rec);
    procedure getPushReq(const CInfo : client_info_rec);
    procedure DeleteClientBySock(Socket: TCustomWinSocket);
    function GetConnectionCount: integer;
    function GetSocketCount: integer;
    procedure SetListenPort(const Value: integer);
    procedure SendMessageToAll(pBuff: Pointer; Size: integer);
    procedure RouteMessageResponse(const CInfo : client_info_rec);
    procedure BroadcastMessage(const CInfo : client_info_rec);
    function IsNewMessage(const CInfo : client_info_rec) : boolean;
  protected
    procedure sockConnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure sockRead(Sender: TObject; Socket: TCustomWinSocket);
    procedure sockDisconnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure sockError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
    procedure sockAccept(Sender: TObject; Socket: TCustomWinSocket);
    procedure SetupNewHdr(var hdr : gnutella_header);
    procedure FreeAllSockets;
    function CreateCInfo(Socket : TCustomWinSocket; const Host : string) : Pclient_info_rec;
    procedure DisposeCInfo(Socket : TCustomWinSocket);
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure Log(const s : string); //publix

    procedure ConnectNewHost(const Host : string);
    procedure ClearGTNStats;
    procedure SendInitToAll;
    procedure SendOneInit(Socket : TCustomWinSocket);
    procedure BeginSearch(const Query : string; MinSpeed : WORD);
    procedure Listen;
    procedure CancelListen;

    property MessagesIn: integer read FMessagesIn;
    property SearchesIn: integer read FSearchesIn;
    property GTNHosts: integer read FGTNHosts;
    property GTNFiles: integer read FGTNFiles;
    property GTNSize: integer read FGTNSize;
    property GTNAvgPing: integer read FGTNAvgPing;
    property ConnectionCount: integer read GetConnectionCount;
    property SocketCount: integer read GetSocketCount;
    property LocalFileCount: integer read FLocalFileCount;
    property LocalFileSize: integer read FLocalFileSize;
  published
    property ConnectionSpeed: integer read FConnectionSpeed write FConnectionSpeed;
    property ListenPort: integer read FListenPort write SetListenPort;
    property LocalIP: string read FLocalIP write FLocalIP;
    property MyTTL: BYTE read FMyTTL write FMyTTL;
    property MaxTTL: BYTE read FMaxTTL write FMaxTTL;
    property SendInitOnConnect: boolean read FSendInitOnConnect write FSendInitOnConnect;
    property OnCatchHost: TStringEvent read FOnCatchHost write FOnCatchHost;
    property OnClientConnected: TClientSockEvent read FOnClientConnected write FOnClientConnected;
    property OnClientCreated: TClientSockEvent read FOnClientCreated write FOnClientCreated;
    property OnClientProtocolConnect: TClientSockEvent read FOnClientProtocolConnect write FOnClientProtocolConnect;
    property OnInitResponse: TInitRespEvent read FOnInitResponse write FOnInitResponse;
    property OnLogStatus: TStringEvent read FOnLogStatus write FOnLogStatus;
    property OnSearchReq: TSearchReqEvent read FOnSearchReq write FOnSearchReq;
    property OnSrchRespBegin: TSrchRespBeginEvent read FOnSrchRespBegin  write FOnSrchRespBegin;
    property OnSrchRespEnd: TSrchRespEndEvent read FOnSrchRespEnd  write FOnSrchRespEnd;
    property OnSrchRespItem: TSrchRespItemEvent read FOnSrchRespItem write FOnSrchRespItem;
    property OnPushRequest: TPushRequest read FOnPushRequest write FOnPushRequest;
    property OnClientDisconnect: TClientSockEvent read FOnClientDisconnect write FOnClientDisconnect;
  end;

function CoCreateGuid(out guid: TGUID): HResult; stdcall; external 'ole32.dll';
function StringFromCLSID(const clsid: TGUID; out psz: PWideChar): HResult; stdcall; external 'ole32.dll';
procedure CoTaskMemFree(pv: Pointer); stdcall; external 'ole32.dll';

function GUIDToString(const ClassID: TGUID): string;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Internet', [TGnuTellaTrans]);
end;

function GUIDToString(const ClassID: TGUID): string;
var
  P: PWideChar;
begin
  StringFromCLSID(ClassID, P);
  Result := P;
  CoTaskMemFree(P);
end;

function PointerToStr(P : Pointer; iSize : integer) : string;
begin
  Result := '';
  while iSize > 0 do begin
    if iSize > 1 then
      AppendStr(Result, Format('%2.2x ', [PBYTE(P)^]))
    else
      AppendStr(Result, Format('%2.2x', [PBYTE(P)^]));

    dec(iSize);
    P := Pointer(Integer(P) + 1);
  end;
end;

function BytesToStr(ar : array of byte) : string;
var
  I:    integer;
begin
  Result := '';
  for I := low(ar) to high(ar) do
    if I < high(ar) then
      AppendStr(Result, Format('%2.2x ', [ar[i]]))
    else
      AppendStr(Result, Format('%2.2x', [ar[i]]))
end;


{ TGnuTellaTrans }

function TGnuTellaTrans.FuncToStr(Func : BYTE) : string;
begin
  case Func of
    $00:  Result := 'Init';
    $01:  Result := 'Init response';
    $40:  Result := 'Client push req';
    $80:  Result := 'Query';
    $81:  Result := 'Query Response';
    else
      Result := '???';
  end;
end;

procedure TGnuTellaTrans.ClearGTNStats;
begin
  FGTNAvgPing := 0;
  FGTNHosts := 0;
  FGTNSize := 0;
  FGTNFiles := 0;
end;

procedure TGnuTellaTrans.ConnectNewHost(const Host: string);
var
  tmpSocket:  TClientSocket;
  pCInfo:     Pclient_info_rec;
  iPos:       integer;
begin
  if Host = '' then
    exit;

  tmpSocket := TClientSocket.Create(Self);
  SockList.Add(tmpSocket);
  pCInfo := CreateCInfo(tmpSocket.Socket, Host);
  pCInfo^.Incoming := false;

  iPos := Pos(':', Host);
  if iPos = 0 then
    iPos := Length(Host) + 1;

  tmpSocket.Host := copy(Host, 1, iPos - 1);
  tmpSocket.Port := StrToIntDef(copy(Host, iPos + 1, Length(Host)), 6346);

  Log('Host is: ' + tmpSocket.Host + ':' + IntToStr(tmpSocket.Port));
  //Log('Host is: ' + tmpSocket.Host);
  //Log('Port is: ' + IntToStr(tmpSocket.Port));

  with tmpSocket do begin
    OnConnect := sockConnect;
    OnDisconnect := sockDisconnect;
    OnError := sockError;
    OnRead := sockRead;
    Open;
    end;

  if Assigned(FOnClientCreated) then
    FOnClientCreated(pCInfo^);
end;

constructor TGnuTellaTrans.Create(AOwner: TComponent);
begin
  inherited;

  FRouteMsgList := TGnuTelMessageList.Create;

  SockList := TList.Create;
  FMyTTL := 7;
  FMaxTTL := 7;
  FListenPort := 6346;
  FConnectionSpeed := 28;

  ServerSock := TServerSocket.Create(Self);
  ServerSock.Port := FListenPort;
  ServerSock.OnClientConnect := sockConnect;
  ServerSock.OnClientDisconnect := sockDisconnect;
  ServerSock.OnClientError := sockError;
  ServerSock.OnClientRead := sockRead;
  ServerSock.OnAccept := sockAccept;
end;

destructor TGnuTellaTrans.Destroy;
begin
  CancelListen;
  FreeAllSockets;
  SockList.Free;
  FRouteMsgList.Free;
  
  inherited;
end;

procedure TGnuTellaTrans.GetGNUTELLAOK(Socket: TCustomWinSocket);
var
  pCInfo: Pclient_info_rec;
  iBytes: integer;
begin
  pCInfo := Pclient_info_rec(Socket.Data);
    { 13 is for GNUTELLA OKaa }
  pCInfo^.RecvBuf.Size := 13;

  iBytes := Socket.ReceiveBuf(pCInfo^.RecvBuf.Memory^,
    pCInfo^.RecvBuf.Size - pCInfo^.RecvBuf.Position);
  if iBytes = SOCKET_ERROR then begin
    Log('Socket error in GNUTELLA OK');
    Socket.Close;
  end
  else begin
    pCInfo^.RecvBuf.Position := pCInfo^.RecvBuf.Position + iBytes;
    if pCInfo^.RecvBuf.Position = pCInfo^.RecvBuf.Size then
      if StrLComp(PChar('GNUTELLA OK'#10#10), PChar(pCInfo^.RecvBuf.Memory),
        pCInfo^.RecvBuf.Size) = 0 then begin
        pCInfo^.RecvBuf.Clear;
        pCInfo.ProtocolConnected := true;
        if Assigned(FOnClientProtocolConnect) then
          FOnClientProtocolConnect(pCInfo^);
        if FSendInitOnConnect then
          SendOneInit(Socket);
      end  // we got GNUTELLA OK

      else begin
        Log('Got invalid GNUTELLA response');
        Socket.Close;
      end;  // not GNUTELLA OK
  end;  // if no SOCKET_ERROR
end;

procedure TGnuTellaTrans.Log(const s: string);
begin
  if Assigned(FOnLogStatus) then
    FOnLogStatus(Self, s);
end;

procedure TGnuTellaTrans.ReadSocketData(Socket: TCustomWinSocket);
var
  Buf:    array[0..8195] of char;
  pCInfo: Pclient_info_rec;
  iBytes: integer;
begin
  pCInfo := Pclient_info_rec(Socket.Data);

  repeat
    iBytes := Socket.ReceiveBuf(Buf, sizeof(Buf));

    if iBytes <> SOCKET_ERROR then
      DealWithData(pCInfo^, @Buf, iBytes);
  until iBytes < sizeof(Buf);
end;

procedure TGnuTellaTrans.SendMessageToAll(pBuff : Pointer; Size : integer);
var
  I:    integer;
begin
  for I := 0 to SockList.Count - 1 do
    if Pclient_info_rec(TClientSocket(SockList[I]).Socket.Data)^.ProtocolConnected then
      TClientSocket(SockList[I]).Socket.SendBuf(pBuff^, Size);
  for I := 0 to ServerSock.Socket.ActiveConnections - 1 do
    if Pclient_info_rec(ServerSock.Socket.Connections[I].Data)^.ProtocolConnected then
      ServerSock.Socket.Connections[I].SendBuf(pBuff^, Size);
end;

procedure TGnuTellaTrans.SendInitToAll;
var
  hdr:  gnutella_header;
begin
  SetupNewHdr(hdr);
  hdr.Func := 0;
  SendMessageToAll(@hdr, sizeof(hdr));
end;

procedure TGnuTellaTrans.SetupNewHdr(var hdr: gnutella_header);
begin
  FillChar(hdr, sizeof(hdr), 0);
  CoCreateGUID(hdr.MsgID);
  hdr.TTLRemaining := FMyTTL;
end;

procedure TGnuTellaTrans.sockConnect(Sender: TObject;
  Socket: TCustomWinSocket);
var
  pCInfo: Pclient_info_rec;
begin
    { if socket.data is not assigned, then this is an incoming connection }
  if not Assigned(Socket.Data) then begin
    pCInfo := CreateCInfo(Socket, Format('%s:%d', [Socket.RemoteAddress,
      Socket.LocalPort]));
    pCInfo^.Incoming := true;
    if Assigned(FOnClientCreated) then
      FOnClientCreated(pCInfo^);
  end;

    { if the socket is outboud, we need to init the protocol }
  if not Pclient_info_rec(Socket.Data)^.Incoming then
    Socket.SendText('GNUTELLA CONNECT/0.4'#10#10);

//  Log('Socket Connect');

  if Assigned(FOnClientConnected) then
    FOnClientConnected(Pclient_info_rec(Socket.Data)^);
end;

procedure TGnuTellaTrans.sockDisconnect(Sender: TObject; Socket: TCustomWinSocket);
begin
//  Log('Socket Disconnect');

  if Assigned(Socket.Data) then begin
    if Assigned(FOnClientDisconnect) then
      FOnClientDisconnect(Pclient_info_rec(Socket.Data)^);
    DisposeCInfo(Socket);
    end;

  DeleteClientBySock(Socket);
end;

procedure TGnuTellaTrans.sockError(Sender: TObject;
  Socket: TCustomWinSocket; ErrorEvent: TErrorEvent;
  var ErrorCode: Integer);
begin
  Log('Socket Error ' + IntToStr(ErrorCode));

  if ErrorCode = 10049 then begin
    Log('Host unreachable');
    sockDisconnect(Sender, Socket);
    ErrorCode := 0;
  end
  else if ErrorCode = 10051 then begin
    Log('Network unreachable');
    sockDisconnect(Sender, Socket);
    ErrorCode := 0;
  end
  else if ErrorCode = 10053 then begin
    Log('Connection aborted');
    sockDisconnect(Sender, Socket);
    ErrorCode := 0;
  end
  else if ErrorCode = 10060 then begin
    Log('Connection timed out');
    sockDisconnect(Sender, Socket);
    ErrorCode := 0;
  end
  else if ErrorCode = 10061 then begin
    Log('Connection refused');
    sockDisconnect(Sender, Socket);
    ErrorCode := 0;
  end;
  if ErrorCode = 10065 then begin
    Log('Host unreachable');
    sockDisconnect(Sender, Socket);
    ErrorCode := 0;
  end
end;

procedure TGnuTellaTrans.sockRead(Sender:TObject; Socket:TCustomWinSocket);
begin
    { if we haven't set our own IP, this is the time to do it }
  if FLocalIP = '' then
    if Socket.LocalAddress <> '127.0.0.1' then
      FLocalIP := Socket.LocalAddress;

  if not Pclient_info_rec(Socket.Data)^.ProtocolConnected then
    if Pclient_info_rec(Socket.Data)^.Incoming then
      GetGNUTELLA_CONNECT(Socket)
    else
      GetGNUTELLAOK(Socket)
  else
    ReadSocketData(Socket);
end;

procedure TGnuTellaTrans.SendOneInit(Socket: TCustomWinSocket);
var
  hdr:    gnutella_header;
begin
  SetupNewHdr(hdr);
  hdr.Func := 0;
  Socket.SendBuf(hdr, sizeof(hdr));
end;

procedure TGnuTellaTrans.DealWithData(CInfo: client_info_rec; pData: PChar;
  DataSize: integer);
var
  iBytesNeeded: integer;
  iBytesDealt:  integer;
begin
  iBytesDealt := 0;
  while iBytesDealt < DataSize do begin
    iBytesNeeded := sizeof(gnutella_header) + CInfo.RecvBuf.Position;
      { check for an underflow }
    if iBytesNeeded > DataSize - iBytesDealt then
      iBytesNeeded := DataSize - iBytesDealt;

    CInfo.RecvBuf.Write(PChar(Integer(pData) + iBytesDealt)^, iBytesNeeded);
    inc(iBytesDealt, iBytesNeeded);

      { we don't have enough data for to be a header }
    if CInfo.RecvBuf.Position < sizeof(gnutella_header) then
      break;

    iBytesNeeded := Pgnutella_header(CInfo.RecvBuf.Memory)^.DataLen;

    if (iBytesNeeded > 3000)
    or (iBytesNeeded < 0) then begin
      Log('I think I''ve got a bogus packet:');
      hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);
      CInfo.RecvBuf.Clear;
      break;
      end;
      { check for an underflow }
    if iBytesNeeded > DataSize - iBytesDealt then
      iBytesNeeded := DataSize - iBytesDealt;
    CInfo.RecvBuf.Write(PChar(Integer(pData) + iBytesDealt)^, iBytesNeeded);
    inc(iBytesDealt, iBytesNeeded);

      { we don't have the whole message yet }
    if CInfo.RecvBuf.Position <
      (sizeof(gnutella_header) + Pgnutella_header(CInfo.RecvBuf.Memory)^.DataLen) then
      break;

    if IsNewMessage(CInfo) then
      GotNewMessage(CInfo);
    CInfo.RecvBuf.Clear;
  end;  { while Bytes }
end;

procedure TGnuTellaTrans.hdr_to_text(const hdr: gnutella_header);
begin
{ BRY!
if hdr.Func in [$00, $01, $40, $80, $81] then
exit;
}
  with hdr do begin
    //Log('Message ID: ' + GUIDToString(MsgID));
    Log('ID: ' //0x' + Format('%2.2x', [Func])
      //+ ' (' + FuncToStr(Func) + ')'
      + FuncToStr(Func)
      + ' TTL: ' + IntToStr(TTLRemaining)
      + ' Hops: ' + IntToStr(HopsTaken)
      + ' Len: ' + IntToStr(DataLen)
      + ' ' + GUIDToString(MsgID)
      );
    end;

//  Log('Function ID: 0x' + Format('%2.2x', [hdr.Func]) + ' (' + FuncToStr(hdr.Func) + ')');
//  Log('TTL Remaining: ' + IntToStr(hdr.TTLRemaining));
//  Log('Hops Taken: ' + IntToStr(hdr.HopsTaken));
//  Log('Data Len: ' + IntToStr(hdr.DataLen));
end;

procedure TGnuTellaTrans.GotNewMessage(const CInfo: client_info_rec);
begin
  if (CInfo.RecvBuf=nil)
  or (CInfo.RecvBuf.Memory=nil)
  or (CInfo.RecvBuf.size<4) then
    exit;

  //hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);

//exit;
    { TODO:  Hmmm.  We need to see if this message is destined for us... }
  if Pgnutella_header(CInfo.RecvBuf.Memory)^.Func and $01 = $01 then
    RouteMessageResponse(Cinfo)
  else
    BroadcastMessage(CInfo);

  case Pgnutella_header(CInfo.RecvBuf.Memory)^.Func of
    $00:  begin
//          hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);
          getInit(CInfo);
          end;
    $01:  begin
          hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);
          getInitResponse(CInfo);
          end;
    $40:  begin
//          hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);
          getPushReq(CInfo);
          end;
    $80:  begin
//          hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);
          getQuery(CInfo);
          end;
    $81:  begin
          hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);
          getQueryResponse(CInfo);
          end;
    else begin
      hdr_to_text(Pgnutella_header(CInfo.RecvBuf.Memory)^);
      getUnknown(CInfo);
      end;
  end;

  inc(FMessagesIn);
end;

procedure TGnuTellaTrans.getInit(const CInfo: client_info_rec);
type
  gir = packed record
    hdr:  gnutella_header;
    ir:   gnutella_init_response;
  end;
var
  gir_out:  gir;
begin
//  Log('Init message in, sending response...');
exit;
  fillchar(gir_out, sizeof(gir_out), 0);
  move(CInfo.RecvBuf.Memory^, gir_out.hdr, sizeof(gnutella_header));
  gir_out.hdr.Func := 1;
  gir_out.hdr.DataLen := sizeof(gir_out.ir);
  gir_out.ir.Port := htons(FListenPort);
  gir_out.ir.ip := inet_addr(PChar(FLocalIP));
  gir_out.ir.filecnt := FLocalFileCount;
  gir_out.ir.totsize := FLocalFileSize;
  Cinfo.Socket.SendBuf(gir_out, sizeof(gir_out));
end;

procedure TGnuTellaTrans.getInitResponse(const CInfo: client_info_rec);
var
  pir:   Pgnutella_init_response;
  ia:   in_addr;
  sAddr:  string;
begin
//  Log('Init response:');

  pir := Pgnutella_init_response(
    integer(CInfo.RecvBuf.Memory) + sizeof(gnutella_header));

  inc(FGTNHosts);
  inc(FGTNFiles, pir^.FileCnt);
    { totsize is in KB, we want MB }
  inc(FGTNSize, pir^.totsize div 1024);

//  Log('  Remote Port: ' + IntToStr(pir^.Port));
//  ia.S_addr := pir^.ip;
//  sAddr := inet_ntoa(ia);
//  Log('  Remote IP: ' + sAddr);
//  Log('  Files shared: ' + IntToStr(pir^.filecnt));
//  Log('  Total Size: ' + IntToStr(pir^.totsize) + ' KB');

  with pir^ do begin
    ia.S_addr := IP;
    sAddr := inet_ntoa(ia);
    Log('Init: ' + sAddr +':'+ IntToStr(Port)
       +' Files: ' + IntToStr(filecnt)
       +' Size: ' + IntToStr(totsize) + ' KB');
    end;

  if Assigned(FOnCatchHost) then
    FOnCatchHost(Self, Format('%s:%d', [sAddr, pir^.Port]));

  if Assigned(FOnInitResponse) then
    FOnInitResponse(sAddr, pir^.Port, pir^.filecnt, pir^.totsize);
end;

procedure TGnuTellaTrans.getPushReq(const CInfo: client_info_rec);
var
  ppr:    Pgnutella_push_req;
  sAddr:  string;
  ia:     in_addr;
begin
//  Log('Client Push request:');

  ppr := Pgnutella_push_req(
    Integer(CInfo.RecvBuf.Memory) + sizeof(gnutella_header));

//  Log('  ClientID128: ' + GUIDToString(ppr^.ClientID128));
//  Log('  FileIndex: ' + IntToStr(ppr^.FileIndex));
  ia.S_addr := ppr^.RequestIP;
  sAddr := string(inet_ntoa(ia));
//  Log('  Requester IP: ' + sAddr);
//  Log('  Requester Port: ' + IntToStr(ppr^.RequestPort));

  if Assigned(FOnCatchHost) then
    FOnCatchHost(Self, Format('%s:%d', [sAddr, ppr^.RequestPort]));

  if Assigned(FOnPushRequest) then
    FOnPushRequest(Self,ppr);
end;

procedure TGnuTellaTrans.getQuery(const CInfo: client_info_rec);
var
  wMinSpeed:  WORD;
  sSearch:    string;
begin
  wMinSpeed := PWORD(Integer(CInfo.RecvBuf.Memory) + sizeof(gnutella_header))^;
  sSearch := string(PChar(
    Integer(CInfo.RecvBuf.Memory) + sizeof(gnutella_header) + 2));

  inc(FSearchesIn);

//  Log('Query:');
//  Log('Minimum Speed: ' + IntToStr(wMinSpeed));
//  Log('Search Criteria: ' + sSearch);

  if Assigned(FOnSearchReq) then
    FOnSearchReq(wMinSpeed, sSearch);
end;

procedure TGnuTellaTrans.getQueryResponse(const CInfo: client_info_rec);
var
  pqrh:   Pgnutella_query_response_hdr;
  ia:     in_addr;
  P:      Pointer;
  PinP:   PChar;
  I:      integer;
  s:      string;
  iFileSize:integer;
  iFileIdx: integer;
begin
  Log('Query Response:');
  pqrh := Pgnutella_query_response_hdr(
    integer(CInfo.RecvBuf.Memory) + sizeof(gnutella_header));

  Log('  Remote Port: ' + IntToStr(pqrh^.port));
  ia.S_addr := pqrh^.ip;
  s := inet_ntoa(ia);
  Log('  Remote IP: ' + s);
  Log('  Remote Speed: ' + IntToStr(pqrh^.speed));

  if Assigned(FOnCatchHost) then
    FOnCatchHost(Self, Format('%s:%d', [s, pqrh^.port]));

  if Assigned(FOnSrchRespBegin) then
    FOnSrchRespBegin(s, pqrh^.port, pqrh^.speed);

  P := Pointer(integer(CInfo.RecvBuf.Memory) + sizeof(gnutella_header)
    + sizeof(gnutella_query_response_hdr));
  PinP := P;
  for I := 1 to pqrh^.nrecs do begin
    iFileIdx := Pgnutella_query_response_rec_hdr(PinP)^.FileIdx;
    Log('    File Index: ' + IntToStr(iFileIdx));
    iFileSize := Pgnutella_query_response_rec_hdr(PinP)^.FileSize;
    Log('    File Size: ' + IntToStr(iFileSize));
    inc(PinP, sizeof(gnutella_query_response_rec_hdr));
    s := string(PChar(PinP));
    Log('    File Name: ' + s);

    inc(PinP, Length(s) + 2);
    if Assigned(FOnSrchRespItem) then
      FOnSrchRespItem(iFileIdx, iFileSize, pqrh^.speed, s);
  end;  { for Recs }

  Log('  ClientID128: ' + GUIDToString(
    Pgnutella_query_response_ftr(PinP)^.clientid128
  ));

  if Assigned(FOnSrchRespEnd) then
    FOnSrchRespEnd(Pgnutella_query_response_ftr(PinP)^.clientid128);
end;

procedure TGnuTellaTrans.getUnknown(const CInfo: client_info_rec);
begin
  Log('Unknown Data: ');

  Log(#13#10 + PointerToStr(
    PChar(Integer(CInfo.RecvBuf.Memory) + sizeof(gnutella_header)),
    Pgnutella_header(CInfo.RecvBuf.Memory)^.DataLen));
end;

procedure TGnuTellaTrans.DeleteClientBySock(Socket: TCustomWinSocket);
var
  I:    integer;
begin
  for I := 0 to SockList.Count - 1 do
    if TClientSocket(SockList[I]).Socket = Socket then begin
      TClientSocket(SockList[I]).Free;
      SockList.Delete(I);
      exit;
    end;
end;


procedure TGnuTellaTrans.BeginSearch(const Query: string; MinSpeed: WORD);
var
  hdr:  gnutella_header;
  P:    PChar;
  I:    integer;
begin
  SetupNewHdr(hdr);
  hdr.Func := $80;
  hdr.DataLen := 2 + Length(Query) + 1;
  GetMem(P, sizeof(Hdr) + 2 + Length(Query) + 1);
  try
    FillChar(P^, sizeof(Hdr) + 2 + Length(Query) + 1, 0);
    Move(hdr, P^, sizeof(hdr));
    Move(MinSpeed, Pointer(Integer(P) + sizeof(Hdr))^, sizeof(MinSpeed));
    StrPCopy(PChar(Integer(P) + sizeof(Hdr) + 2), PChar(Query));
    for I := 0 to SockList.Count - 1 do
      if Pclient_info_rec(TClientSocket(SockList[I]).Socket.Data)^.ProtocolConnected then
        TClientSocket(SockList[I]).Socket.SendBuf(P^, sizeof(Hdr) + 2 + Length(Query) + 1);
  finally
    FreeMem(P);
  end;
end;

function TGnuTellaTrans.GetConnectionCount: integer;
var
  I:    integer;
begin
  Result := 0;
  for I := 0 to SockList.Count - 1 do
    if Pclient_info_rec(TClientSocket(SockList[I]).Socket.Data)^.ProtocolConnected then
      inc(Result);
  for I := 0 to ServerSock.Socket.ActiveConnections - 1 do
    if Pclient_info_rec(ServerSock.Socket.Connections[I].Data)^.ProtocolConnected then
      inc(Result);
end;

function TGnuTellaTrans.GetSocketCount: integer;
begin
  Result := SockList.Count;
end;

procedure TGnuTellaTrans.FreeAllSockets;
var
  I:    integer;
begin
  for I := 0 to SockList.Count - 1 do begin
    DisposeCInfo(TClientSocket(SockList[I]).Socket);
    TClientSocket(SockList[I]).Free;
  end;

  SockList.Clear;
end;

procedure TGnuTellaTrans.DisposeCInfo(Socket : TCustomWinSocket);
begin
  Pclient_info_rec(Socket.Data)^.RecvBuf.Free;
  Dispose(Pclient_info_rec(Socket.Data));
  Socket.Data := nil;
end;

procedure TGnuTellaTrans.SetListenPort(const Value: integer);
begin
  if FListenPort = Value then
    exit;

  if ServerSock.Active then
    raise Exception.Create('Setting port while server open not yet supported'); 

  FListenPort := Value;
  ServerSock.Port := FListenPort;
end;

procedure TGnuTellaTrans.GetGNUTELLA_CONNECT(Socket: TCustomWinSocket);
var
  pCInfo: Pclient_info_rec;
  iBytes: integer;
begin
  pCInfo := Pclient_info_rec(Socket.Data);
    { 22 is for GNUTELLA CONNECT/0.4aa }
  pCInfo^.RecvBuf.Size := 22;

  iBytes := Socket.ReceiveBuf(pCInfo^.RecvBuf.Memory^,
    pCInfo^.RecvBuf.Size - pCInfo^.RecvBuf.Position);
  if iBytes = SOCKET_ERROR then begin
    Log('Socket error in GNUTELLA CONNECT/0.4');
    Socket.Close;
  end
  else begin
    pCInfo^.RecvBuf.Position := pCInfo^.RecvBuf.Position + iBytes;
    if pCInfo^.RecvBuf.Position = pCInfo^.RecvBuf.Size then
      if StrLComp(PChar('GNUTELLA CONNECT/0.4'#10#10), PChar(pCInfo^.RecvBuf.Memory),
        pCInfo^.RecvBuf.Size) = 0 then begin
        pCInfo^.RecvBuf.Clear;
        Socket.SendText('GNUTELLA OK'#10#10);
        pCInfo.ProtocolConnected := true;
        if Assigned(FOnClientProtocolConnect) then
          FOnClientProtocolConnect(pCInfo^);
        if FSendInitOnConnect then
          SendOneInit(Socket);
      end  // we got GNUTELLA OK

      else begin
        Log('Got invalid GNUTELLA protocol init');
        Socket.Close;
      end;  // not GNUTELLA OK
  end;  // if no SOCKET_ERROR
end;

procedure TGnuTellaTrans.CancelListen;
var
  I:    integer;
begin
  for I := 0 to ServerSock.Socket.ActiveConnections - 1 do
    ServerSock.Socket.Connections[I].Close;
    
  ServerSock.Close;
end;

procedure TGnuTellaTrans.Listen;
begin
  ServerSock.Open;
end;

procedure TGnuTellaTrans.sockAccept(Sender: TObject;
  Socket: TCustomWinSocket);
begin
end;

function TGnuTellaTrans.CreateCInfo(Socket: TCustomWinSocket;
  const Host: string): Pclient_info_rec;
begin
  New(Result);
  Result^.Socket := Socket;
  Socket.Data := Result;
  Result^.Host := Host;
  Result^.ProtocolConnected := false;
  Result^.RecvBuf := TMemoryStream.Create;
end;

procedure TGnuTellaTrans.BroadcastMessage(const CInfo: client_info_rec);
begin
  { TODO:  Add the messageid to the route list, and send to everyone except CInfo.Socket }
;
end;

procedure TGnuTellaTrans.RouteMessageResponse(
  const CInfo: client_info_rec);
begin
  { TODO:  Find the message in the route list, and send the response to only that CInfo.Socket }
;
end;

function TGnuTellaTrans.IsNewMessage(const CInfo: client_info_rec): boolean;
var
  iPos:   integer;
begin
    { response messages are always considered new }
  if Pgnutella_header(CInfo.RecvBuf.Memory)^.Func and $01 = $01 then begin
    Result := true;
    exit;
  end;

  Result := not FRouteMsgList.Find('' { CInfo.MessageID }, iPos);
end;

{ TGnuTelMessageList }

function TGnuTelMessageList.Find(const S: string;
  var Index: Integer): Boolean;
begin
  Result := false;
end;

end.
